from typing import Dict
from asciitree.drawing import BoxStyle, BOX_DOUBLE
from asciitree import LeftAligned
from collections import OrderedDict as OD

from baobao import Pipeline, Step
from baobao.pipeline import str_


def print_pipeline(pipeline: Pipeline) -> None:
    def tree_pipeline(pipeline: pipeline) -> Dict:
        return {
            f"pipeline << {pipeline.root_node.__name__}": OD([
                (f"Step {ii}:{step.func.__name__}", tree_step(step))
                for ii, step in enumerate(pipeline.pipeline)
            ])}

    def tree_step(step: Step) -> Dict:
        return {
            f"{kk}{'' if isinstance(vv, Pipeline) else ' = ' + str_(vv)}":
                tree_pipeline(vv) if isinstance(vv, Pipeline) else {}
            for kk, vv in {
                **{str(ii): arg for ii, arg in enumerate(step.args)},
                **step.kwargs
            }.items()
        }

    print(LeftAligned(
        draw=BoxStyle(gfx=BOX_DOUBLE, horiz_len=2)
    )(tree_pipeline(pipeline)))



from multiprocessing import Pool
from multiprocessing.pool import ApplyResult
from typing import Union
import logging

from ..pipeline import Pipeline, Step


def run_parallel(pipeline: Pipeline, n_jobs: int = 2):
    def extraxt_pipelines_from_step(step: Step):
        return [extract_pipelines_from_pipeline(pp) for pp in (
                [arg for arg in step.args if isinstance(arg, Pipeline)] +
                [arg for _, arg in step.kwargs.items() if isinstance(arg, Pipeline)]
        )]

    def extract_pipelines_from_pipeline(pipeline: Pipeline):
        return [extraxt_pipelines_from_step(step) for step in pipeline.pipeline]

    def parse_pipeline(pipeline: Pipeline) -> Pipeline:
        pipeline.pipeline = [parse_step(step) for step in pipeline.pipeline]
        return pipeline

    def parse_arg(arg):
        return itter_or_submitt_pipeline(arg) if isinstance(arg, Pipeline) else arg

    def itter_or_submitt_pipeline(pipeline: Pipeline) -> Union[Pipeline, ApplyResult]:
        if len(extract_pipelines_from_pipeline(pipeline)) == 0:
            logging.info(f"Submit: {str(pipeline)} to: {str(pl)}")
            return pl.apply_async(pipeline)
        else:
            return parse_pipeline(pipeline)

    def parse_step(step):
        step.arg = [parse_arg(arg) for arg in step.args]
        step.kwargs = {k: parse_arg(v) for k, v in step.kwargs.items()}
        return step

    with Pool(n_jobs) as pl:
        return parse_pipeline(pipeline).run()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['baobao', 'baobao.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'baobao',
    'version': '0.2.1',
    'description': '',
    'long_description': '![image.png](notebooks/baobao_logo.png)\n\nBaobao is a small library that attempts writing and maintaining data pipelines easier. It was written with Pandas in mind, but is not limited to it or even depends on it. It\'s AIP closely models Pandas pipe functionality.\n\nDesign principles are:\n- Simple in the core\n- Minimal dependencies\n- Additional functionality is optional\n\nPipelines can be build like this:\n\n\n```python\nimport pandas as pd\nfrom baobao import Pipeline  # To define pipelines\nfrom baobao import Step  # Pipelines are build from Steps which may contain Pipelines\n\n# some functions that return pd.Dataframes to play with\nfrom baobao.utils.examples import load_1, load_2, load_3\n\nPipeline(\n    root_node=load_1,  # We have to start somewhere, could be any object or callable\n    pipeline=(  # Pipelines are build from Steps\n        Step(\n            # Each Step is build from a function that takes \n            # the output of the previous Step as input in \n            # the first argument\n            func=pd.merge,\n            # Any argoments to [func] can be given\n            left_index=True, right_index=True,\n            right=Pipeline(load_2),  # including additional Pipelines\n        ),\n        Step(func=pd.merge, right=Pipeline(load_3), left_index=True, right_index=True),\n    )\n).run()  # Call the run method to actually run the pipeline, enjoy some logging out of the box\n```\n    INFO:root:Pipeline(root_node=load_1(),memory=None)\n    INFO:root:    Step(func=merge,args=(),left_index=True,right_index=True,right=Pipeline(root_node=l[...])\n    INFO:root:Pipeline(root_node=load_2(),memory=None)\n    Load 1\n    INFO:root:Complete pipeline after 3.02s\n    INFO:root:    Step(func=merge,args=(),right=Pipeline(root_node=l[...],left_index=True,right_index=True)\n    INFO:root:Pipeline(root_node=load_3(),memory=None)\n    Load 2\n    INFO:root:Complete pipeline after 3.02s\n    INFO:root:Complete pipeline after 9.06s\n    Load 3\n\n\n\nA neat way of speeding this up is caching results to disk:\n\n\n```python\n# Baobao doesn\'t depend on any caching but respects Joblibs \n# API in case you want to chose some other caching strategy  \nfrom joblib import Memory\n\n# Baobao options are separated into a dedicated object in\n# order to make it simple to push options down to Pipelines\n# included in Steps of the root Pipeline\nfrom baobao import PipelineOpts\n\n# Define the pipeline:\npipeline = Pipeline(\n    root_node=load_1,\n    opts=PipelineOpts(\n        memory=Memory("./cache", verbose=0),\n        push_options=True  # Pushing options down to included Pipelines\n    ),\n    pipeline=(\n        Step(func=pd.merge, right=Pipeline(load_2), left_index=True, right_index=True),\n        Step(func=pd.merge, right=Pipeline(load_3), left_index=True, right_index=True),\n    )\n)\n# Run the pipeline:\npipeline.run()\n```\n\n    INFO:root:Pipeline(root_node=load_1(),memory=Memory(location=./cache/joblib))\n    INFO:root:    Step(func=merge,args=(),right=Pipeline(root_node=l[...],left_index=True,right_index=True)\n    INFO:root:    Pipeline(root_node=load_2(),memory=Memory(location=./cache/joblib))\n    INFO:root:    Complete pipeline after 0.00s\n    INFO:root:    Step(func=merge,args=(),right=Pipeline(root_node=l[...],left_index=True,right_index=True)\n    INFO:root:    Pipeline(root_node=load_3(),memory=Memory(location=./cache/joblib))\n    INFO:root:    Complete pipeline after 0.01s\n    INFO:root:Complete pipeline after 0.04s\n\nReasoning over pipelines might be simpler with this printing utility:\n\n\n```python\nfrom baobao.utils import print_pipeline\nprint_pipeline(pipeline)\n```\n\n    pipeline << load_1\n     ╠══ Step 0:merge\n     ║   ╠══ right = DataFrame shape:(10, 1), columns:[\'c2\']\n     ║   ╠══ left_index = True\n     ║   ╚══ right_index = True\n     ╚══ Step 1:merge\n         ╠══ right = DataFrame shape:(10, 1), columns:[\'c3\']\n         ╠══ left_index = True\n         ╚══ right_index = True\n\n\nPrinting can be adjusted using the multiple dispatch pattern:\n\n\n```python\nfrom baobao.pipeline import str_\n@str_.register\ndef _(inp: pd.DataFrame):\n    return f"DataFrame(shape:{inp.shape})"\n@str_.register\ndef _(inp: bool):\n    return f"Bool:{inp}"\n\nprint_pipeline(pipeline)\n```\n\n    pipeline << load_1\n     ╠══ Step 0:merge\n     ║   ╠══ right = DataFrame(shape:(10, 1))\n     ║   ╠══ left_index = Bool:True\n     ║   ╚══ right_index = Bool:True\n     ╚══ Step 1:merge\n         ╠══ right = DataFrame(shape:(10, 1))\n         ╠══ left_index = Bool:True\n         ╚══ right_index = Bool:True\n\n\nA little more depth added:\n\n\n```python\nfrom baobao.utils.examples import *  # Import more load_X functions\n\ndef mk_pipeline():  # get fresh pipeline each time we call this function\n    return Pipeline(\n        root_node=load_1,\n        opts=PipelineOpts(\n            memory=None,#Memory("./cache", verbose=0),\n            push_options=True\n        ),\n        pipeline=(\n            Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(load_2)),\n            Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(\n                root_node=load_3,\n                pipeline=(\n                    Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(load_4)),\n                    Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(\n                        root_node=load_2,\n                        pipeline=(\n                            Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(load_5)),\n                            Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(load_6)),\n                        )\n                    )),\n                ))\n                ),\n            Step(func=pd.merge, left_index=True, right_index=True, right=Pipeline(load_5)),\n        )\n    )\n```\n\nLets print this one again:\n\n\n```python\nprint_pipeline(mk_pipeline())\n```\n\n    pipeline << load_1\n     ╠══ Step 0:merge\n     ║   ╠══ left_index = Bool:True\n     ║   ╠══ right_index = Bool:True\n     ║   ╚══ right\n     ║       ╚══ pipeline << load_2\n     ╠══ Step 1:merge\n     ║   ╠══ left_index = Bool:True\n     ║   ╠══ right_index = Bool:True\n     ║   ╚══ right\n     ║       ╚══ pipeline << load_3\n     ║           ╠══ Step 0:merge\n     ║           ║   ╠══ left_index = Bool:True\n     ║           ║   ╠══ right_index = Bool:True\n     ║           ║   ╚══ right\n     ║           ║       ╚══ pipeline << load_4\n     ║           ╚══ Step 1:merge\n     ║               ╠══ left_index = Bool:True\n     ║               ╠══ right_index = Bool:True\n     ║               ╚══ right\n     ║                   ╚══ pipeline << load_2\n     ║                       ╠══ Step 0:merge\n     ║                       ║   ╠══ left_index = Bool:True\n     ║                       ║   ╠══ right_index = Bool:True\n     ║                       ║   ╚══ right\n     ║                       ║       ╚══ pipeline << load_5\n     ║                       ╚══ Step 1:merge\n     ║                           ╠══ left_index = Bool:True\n     ║                           ╠══ right_index = Bool:True\n     ║                           ╚══ right\n     ║                               ╚══ pipeline << load_6\n     ╚══ Step 2:merge\n         ╠══ left_index = Bool:True\n         ╠══ right_index = Bool:True\n         ╚══ right\n             ╚══ pipeline << load_5\n\n\nand run it in a sequential manner:\n\n\n```python\nmk_pipeline().run()\n```\n    INFO:root:Pipeline(root_node=load_1(),memory=None)\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:    Pipeline(root_node=load_2(),memory=None)\n    Load 1\n    INFO:root:    Complete pipeline after 3.01s\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:    Pipeline(root_node=load_3(),memory=None)\n    Load 2\n    INFO:root:        Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:        Pipeline(root_node=load_4(),memory=None)\n    Load 3\n    INFO:root:        Complete pipeline after 3.02s\n    INFO:root:        Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:        Pipeline(root_node=load_2(),memory=None)\n    Load 4\n    INFO:root:            Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:            Pipeline(root_node=load_5(),memory=None)\n    Load 2\n    INFO:root:            Complete pipeline after 3.01s\n    INFO:root:            Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:            Pipeline(root_node=load_6(),memory=None)\n    Load 5\n    INFO:root:            Complete pipeline after 3.01s\n    INFO:root:        Complete pipeline after 9.05s\n    INFO:root:    Complete pipeline after 15.10s\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:    Pipeline(root_node=load_5(),memory=None)\n    Load 6\n    INFO:root:    Complete pipeline after 3.01s\n    INFO:root:Complete pipeline after 24.15s\n    Load 5\n\nWe can speed things up by submitting those pipelines that do not contain further pipelines to a multiprocess Pool and enjoy some nice speedups:\n\n\n```python\nfrom baobao.utils import run_parallel\n\nrun_parallel(mk_pipeline(), n_jobs=4)\n```\n\n    INFO:root:Submit: Pipeline(root_node=load_2(),memory=None) to: <multiprocessing.pool.Pool state=RUN pool_size=4>\n    INFO:root:Submit: Pipeline(root_node=load_4(),memory=None) to: <multiprocessing.pool.Pool state=RUN pool_size=4>\n    INFO:root:Submit: Pipeline(root_node=load_5(),memory=None) to: <multiprocessing.pool.Pool state=RUN pool_size=4>\n    INFO:root:Pipeline(root_node=load_2(),memory=None)\n    INFO:root:Submit: Pipeline(root_node=load_6(),memory=None) to: <multiprocessing.pool.Pool state=RUN pool_size=4>\n    INFO:root:Pipeline(root_node=load_4(),memory=None)\n    INFO:root:Submit: Pipeline(root_node=load_5(),memory=None) to: <multiprocessing.pool.Pool state=RUN pool_size=4>\n    INFO:root:Pipeline(root_node=load_1(),memory=None)\n    INFO:root:Pipeline(root_node=load_5(),memory=None)\n    INFO:root:Pipeline(root_node=load_6(),memory=None)\n    Load 4\n    Load 2\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=<multiprocessing.poo[...])\n    Load 6\n    INFO:root:Complete pipeline after 3.04s\n    Load 5\n    INFO:root:Pipeline(root_node=load_5(),memory=None)\n    INFO:root:Complete pipeline after 3.04s\n    INFO:root:Complete pipeline after 3.06s\n    INFO:root:Complete pipeline after 3.06s\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:    Pipeline(root_node=load_3(),memory=None)\n    Load 1\n    Load 5\n    INFO:root:Complete pipeline after 3.02s\n    INFO:root:        Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=<multiprocessing.poo[...])\n    INFO:root:        Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=Pipeline(root_node=l[...])\n    INFO:root:        Pipeline(root_node=load_2(),memory=None)\n    Load 3\n    INFO:root:            Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=<multiprocessing.poo[...])\n    INFO:root:            Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=<multiprocessing.poo[...])\n    INFO:root:        Complete pipeline after 3.02s\n    INFO:root:    Complete pipeline after 6.04s\n    INFO:root:    Step(func=merge,args=(),left_index=Bool:True,right_index=Bool:True,right=<multiprocessing.poo[...])\n    INFO:root:Complete pipeline after 9.15s\n    Load 2\n',
    'author': 'André Hollstein',
    'author_email': 'andre@dr-hollstein.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from copy import copy
from typing import Iterable, Union, Callable, Optional
from functools import singledispatch
import logging
from time import time
from multiprocessing.pool import ApplyResult

#######################
# optional dependencies
#######################

try:
    from joblib import Memory
except ImportError:
    logging.info("baobao: Missing joblib optional dependency, Step memoization will not work.")

    class Memory:
        pass

try:
    import pandas as pd
except ImportError:
    class pd:
        DataFrame = None


@dataclass
class PipelineOpts:
    memory: Optional[Memory] = None
    in_place: bool = True
    push_options: bool = False
    step_prefix = 4 * " "
    pipeline_prefix = ""


def _run(func: Callable, df: pd.DataFrame, *args, **kwargs):
    return func(df, *args, **kwargs)


class Step:

    def __init__(self, func: Callable, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.pipeline_opts = None

    def __str__(self):

        def str_reduce(inp: str, allowed_length: int = 5) -> str:
            if len(inp) < allowed_length:
                return inp
            else:
                return f"{inp[:allowed_length]}[...]"

        def kv2str(kv: Iterable) -> str:
            return '='.join(map(lambda k: str_reduce(str_(k), allowed_length=20), kv))

        return (
            f"Step(func={self.func.__name__}"
            f",args=({','.join([str_reduce(str_(a)) for a in self.args])})"
            f",{','.join([kv2str(kv) for kv in self.kwargs.items()])})"
        )

    def set_options(self, opts):
        self.pipeline_opts = opts
        return self

    def push_options(self):

        opts = copy(self.pipeline_opts)
        opts.step_prefix = f"    {opts.step_prefix}"
        opts.pipeline_prefix = f"    {opts.pipeline_prefix}"

        def _pp(inp):
            if isinstance(inp, Pipeline):
                inp.opts = opts
                return inp
            else:
                return inp

        self.args = [_pp(arg) for arg in self.args]
        self.kwargs = {k: _pp(v) for k, v in self.kwargs.items()}

    def _run_pipelines_for_args_kwargs(self):

        def _run_if_pipeline(inp):
            return inp.get() if any([isinstance(inp, cl) for cl in (Pipeline, ApplyResult)]) else inp

        self.args = [_run_if_pipeline(arg) for arg in self.args]
        self.kwargs = {k: _run_if_pipeline(v) for k, v in self.kwargs.items()}

    def run(self, df: pd.DataFrame) -> pd.DataFrame:

        self._run_pipelines_for_args_kwargs()
        return (self.pipeline_opts.memory.cache(_run) if self.pipeline_opts.memory is not None else _run)(
            self.func, df, *self.args, **self.kwargs)


class Pipeline:
    def __init__(
            self, root_node: Union[pd.DataFrame, Callable], pipeline: Iterable[Step] = tuple(),
            opts: PipelineOpts = PipelineOpts()
    ):
        self.root_node = root_node
        self.opts = opts
        self.pipeline = pipeline
        self.__name__ = str(self)

    def __str__(self):
        return f"Pipeline(root_node={str_(self.root_node)},memory={str(self.opts.memory)})"

    def run(self):
        return self.__call__()

    def get(self):
        return self.__call__()

    def __call__(self):

        t0 = time()
        logging.info(f"{self.opts.pipeline_prefix}{str(self)}")

        self.pipeline = [step.set_options(self.opts) for step in self.pipeline]

        if self.opts.push_options:
            for step in self.pipeline:
                step.push_options()

        # if memory given and root_node is callable, replace by memorized version
        root_node = (
            self.opts.memory.cache(self.root_node)
            if callable(self.root_node) and self.opts.memory is not None else self.root_node
        )
        # materialize root_node
        root_node = root_node() if callable(root_node) else root_node
        df = root_node if self.opts.in_place is True else root_node.copy()
        for ii, step in enumerate(self.pipeline):
            logging.info(f"{self.opts.step_prefix}{str(step)}")
            df = step.run(df)
        logging.info(f"{self.opts.pipeline_prefix}Complete pipeline after {(time()-t0):.2f}s")

        return df


@singledispatch
def str_(inp):
    if callable(inp):
        return f"{inp.__name__}()"
    else:
        return str(inp)[:40]


@str_.register
def _(inp: pd.DataFrame):
    return f"DataFrame shape:{inp.shape}, columns:{str(list(inp.columns))[:50]}"


@str_.register
def _(inp: bool):
    return f"{inp}"



from time import sleep
import pandas as pd

t_sleep = 3.0

def load_1():
    sleep(t_sleep)
    print("Load 1")
    return pd.DataFrame(data={"c1":list(range(10))})

def load_2():
    sleep(t_sleep)
    print("Load 2")
    return pd.DataFrame(data={"c2":[c**2 for c in range(10)]})

def load_3():
    sleep(t_sleep)
    print("Load 3")
    return pd.DataFrame(data={"c3":[c**2 / 2 for c in range(10)]})

def load_4():
    sleep(t_sleep)
    print("Load 4")
    return pd.DataFrame(data={"c4":[c-1 for c in range(10)]})

def load_5():
    sleep(t_sleep)
    print("Load 5")
    return pd.DataFrame(data={"c5":[c*2 for c in range(10)]})

def load_6():
    sleep(t_sleep)
    print("Load 6")
    return pd.DataFrame(data={"c6":[c*2/4 for c in range(10)]})